/*-------------------------------------------------------------------------
	Copyright (C) 2008-2009 American Megatrends Inc.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
------------------------------------------------------------------------- */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include "i2c-data.h"
#include "i2c-hardware.h"
#include "gpio_vsc.h"
#include "gpio_info.h"
#include <asm-mips/mach-vsc452/vsc452.h>
#include "PDKGPIO.h"

//#define DEBUG
#define LOG_BUS_RECOVERY
#define SKIP_RECOVERY_OF_DEVICES_RECOVERED_FROM_GPIO
#define FAN_SLAVE_ADDR              0x33

extern gpio_dev_t *g_gpio_dev;

/* Prototypes */
int is_i2c_bus_hung(int	bus);
int i2c_bus_recovery(int bus);

void reset_mux_via_gpio(int bus);

bus_recovery_info_T m_bus_recovery_info[BUS_COUNT] =
{
#if (BUS_COUNT >= 1)
	{ ENABLE_RECOVERY, ENABLE_SMBUS_RESET, DISABLE_CLOCK_PULSE, ENABLE_FORCE_STOP, DEFAULT_NUM_PULSES, DEFAULT_FREQ },
#endif
#if (BUS_COUNT >= 2)
	{ ENABLE_RECOVERY, ENABLE_SMBUS_RESET, DISABLE_CLOCK_PULSE, ENABLE_FORCE_STOP, DEFAULT_NUM_PULSES, DEFAULT_FREQ },
#endif
#if (BUS_COUNT >= 3)
	{ ENABLE_RECOVERY, ENABLE_SMBUS_RESET, DISABLE_CLOCK_PULSE, ENABLE_FORCE_STOP, DEFAULT_NUM_PULSES, DEFAULT_FREQ },
#endif
#if (BUS_COUNT >= 4)
	{ ENABLE_RECOVERY, ENABLE_SMBUS_RESET, DISABLE_CLOCK_PULSE, ENABLE_FORCE_STOP, DEFAULT_NUM_PULSES, DEFAULT_FREQ },
#endif
#if (BUS_COUNT >= 5)
	{ ENABLE_RECOVERY, ENABLE_SMBUS_RESET, DISABLE_CLOCK_PULSE, ENABLE_FORCE_STOP, DEFAULT_NUM_PULSES, DEFAULT_FREQ },
#endif
#if (BUS_COUNT >= 6)
	{ ENABLE_RECOVERY, ENABLE_SMBUS_RESET, DISABLE_CLOCK_PULSE, ENABLE_FORCE_STOP, DEFAULT_NUM_PULSES, DEFAULT_FREQ },
#endif
};

#define DO_RECOVERY   0
#define SKIP_RECOVERY 1
/* Skip recovery of devices that are recovered from upper layer using gpio */
static int skip_recovery(int bus) 
{
	/* If arbitration bus */
	if ((bus == I2C_BUS_2)) {
		/* if fan device */
		if (vsc_data_ptr[bus].slave_addr == FAN_SLAVE_ADDR) {
			/* Skip recovery of fans since they are recovered from upper layer using gpio */
			return SKIP_RECOVERY;
		}
	}

	return DO_RECOVERY;
}


static void set_0_to_i2c_pin_bit_control_reg(unsigned long bitmask, int bus)
{
	unsigned long i2c_pin_bit_ctrl;

	i2c_pin_bit_ctrl = i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);

	i2c_pin_bit_ctrl &= bitmask;
	
	i2c_vsc_write_reg(i2c_pin_bit_ctrl, (UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);
}

static void set_1_to_i2c_pin_bit_control_reg(unsigned long bitmask, int bus)
{
	unsigned long i2c_pin_bit_ctrl;
	
	i2c_pin_bit_ctrl = i2c_vsc_read_reg((UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);
	
	i2c_pin_bit_ctrl |= bitmask;
	
	i2c_vsc_write_reg(i2c_pin_bit_ctrl, (UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_CNTRL);
}

static void drive_data_low(int bus)
{
	unsigned long i2c_data_bitmask = 0<<0;
	
	if (bus >= 0 && bus < BUS_COUNT)
		set_0_to_i2c_pin_bit_control_reg(i2c_data_bitmask, bus);
}


static void drive_clock_low(int bus)
{
	unsigned long i2c_clk_bitmask = 0<<1;
	
	if (bus >= 0 && bus < BUS_COUNT)
		set_0_to_i2c_pin_bit_control_reg(i2c_clk_bitmask, bus);
}

static void drive_data_high(int bus)
{
	unsigned long i2c_data_bitmask = 1<<0;

	if (bus >= 0 && bus < BUS_COUNT)
		set_1_to_i2c_pin_bit_control_reg(i2c_data_bitmask, bus);
}

static void drive_clock_high(int bus)
{
	unsigned long i2c_clk_bitmask = 1<<1;
	
	if (bus >= 0 && bus < BUS_COUNT)
		set_1_to_i2c_pin_bit_control_reg(i2c_clk_bitmask, bus);
}

static void vsc_bit_bang_deselect(int bus)
{
	/* Enable I2C interface */
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS
				| I2C_ENABLE_INTERFACE
				| I2C_ENABLE_INTERRUPT
				| I2C_ACKNOWLEDGEMENT),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
}

static void vsc_bit_bang_select(int bus)
{
	/* Disable I2C interface to make it act as GPIO pin */
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS),
				(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);
}

static int perform_slave_recovery(int bus)
{
	int i;
	int pulse_period = 50;
	int num_clock_pulses;

	pulse_period = (int)(500000 / (m_bus_recovery_info[bus].frequency));
	num_clock_pulses = m_bus_recovery_info[bus].num_clock_pulses;


#ifdef ARBITRATION_SUPPORT
	if (bus == I2C_BUS_2) {
		RESTART_SHORT_TIMER(RESTART_TIMER_VALUE);  /* Restart support fpga short timer */	
	}
#endif

	/* See if generate clock pulses option is enabled */        
	if (m_bus_recovery_info[bus].is_generate_clock_pulses_enabled) {
#ifdef DEBUG
		printk(KERN_ERR "I2C%d: Performing bus recovery method: Generate clock pulses\n", bus);
#endif	
		/* Perform bit banging on specified bus */
		while (num_clock_pulses > 0) {
			
			num_clock_pulses -= 1;
	
			/* Make SCL low */ 
			drive_clock_low(bus);
			udelay(pulse_period); 
	
			/* Make SCL high */
			drive_clock_high(bus);
			udelay(pulse_period);
	
			/* If bus is no longer hung, return success */ 
			if ((HIGH == get_sda_status(bus)) && (HIGH == get_clk_status(bus))) { 
				return 0;
			}
		}
	}
	
	/* If force stop option is enabled */
	if (m_bus_recovery_info[bus].is_force_stop_enabled) {
#ifdef DEBUG
		printk(KERN_ERR "I2C%d: Performing bus recovery method: Force stop conditions\n", bus);
#endif
		/* Generate multiple stop conditions */
		for (i=0; i<DEFAULT_NUM_PULSES; i++) {

			/* Make SDA and SCL low */
			drive_data_low(bus);
			drive_clock_low(bus);
			udelay(5);
	
			/* Make SCL high */
			drive_clock_high (bus);
			udelay(2);
	
			/* Make SDA high */
			drive_data_high (bus);
			udelay (5);
		}


		return -1;
	}

	
	/* Otherwise, return success */
	return 0;
}

static int perform_vsc_recovery(int bus)
{
#ifdef DEBUG
	printk(KERN_ERR "I2C%d: Performing bus recovery method: VSC I2C core reset\n", bus);
#endif
	i2c_init_internal_data(bus);
	i2c_vsc_reset(bus);
	i2c_init_hardware(bus);

#if 0 // no need to check bus again because we are going to clock slave device out anyway and return error
	/* Quick check to see if bus is still hung? */
	if ((LOW == get_sda_status(bus)) || (LOW == get_clk_status(bus))) {

		/* If so confirm with more thorough check and if hung return error */
		if (is_i2c_bus_hung(bus)) {
			return 1;
		}
	}

	/* Otherwise return success */
	return 0;
#else
	return 1;
#endif
}

int vsc_i2c_bus_recovery(struct i2c_adapter *i2c_adap)
{
    return i2c_bus_recovery(i2c_adap->nr);
}

int vsc_is_i2c_bus_hung(struct i2c_adapter *i2c_adap)
{
    return is_i2c_bus_hung(i2c_adap->nr);
}

int i2c_bus_recovery(int bus)
{
	unsigned char dev_stat_index;

#ifdef SKIP_RECOVERY_OF_DEVICES_RECOVERED_FROM_GPIO
	if (SKIP_RECOVERY == skip_recovery(bus)) {
		return 0;
	}
#endif
	
	i2c_bus_status_ptr[bus].bus_recovery_cnt++;
	dev_stat_index = i2c_vsc_get_dev_stat_index(bus, vsc_data_ptr[bus].slave_addr, g_mux_chan[bus]);
	i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].recovery_cnt++;
	
#ifdef LOG_BUS_RECOVERY
	printk(KERN_ERR "I2C%d: Performing bus recovery %d to slave 0x%x\n", bus, 
		i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].recovery_cnt,
		i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].dev_slave_addr);
#endif
	/* If VSC I2C bus recovery is enabled */
	if (m_bus_recovery_info[bus].is_smbus_reset_enabled) { 

		/* First, assume we are the culprit and reinitialize the VSC452 I2C core */
		if (0 == perform_vsc_recovery(bus)) {
			return 0; /* Success */
		}
	}

	/* VSC452 I2C core reinit failed.  Let's try recovering from slave device hanging the bus */

	/* We have two methods of slave recovery. If both methods of slave 
 	 * recovery are disabled return success (nothing to do).
 	 * Refer to m_bus_recovery_info table to enable/disable 
 	 * I2C bus recovery features on a per bus bassis. */
	if ((m_bus_recovery_info[bus].is_force_stop_enabled) ||
	    (m_bus_recovery_info[bus].is_generate_clock_pulses_enabled)) {

		/* Set SDA output line so that it's not driven by bit-bang */
		drive_data_high(bus);

		/* Set SCL output line so that it's not driven by bit-bang */
		drive_clock_high(bus);

		/* Set bit bang. Disconnect ourselves from the bus */
		vsc_bit_bang_select(bus);

		/* Perform slave recovery */
		if (0 != perform_slave_recovery(bus)) {

			vsc_bit_bang_deselect(bus);
			/* Every attempt of bus recovery has failed! Perhaps mux is in odd state, try reseting the mux */
			/* For Razor, only Bus 3 and Bus 1 have a mux */
			if ( (bus == I2C_BUS_3) || (bus == I2C_BUS_1) ) {
#ifdef DEBUG
				printk(KERN_ERR "I2C%d: Performing bus recovery method: Reset mux via GPIO\n", bus);
#endif
				reset_mux_via_gpio(bus);
			}
			return -1;

		} 

		vsc_bit_bang_deselect(bus);
	}

	/* Return success */
	return 0;
}

int vsc_set_recovery_info(int bus, bus_recovery_info_T* info)
{
	if (bus >= BUS_COUNT) 
		return -1;

	memcpy(&m_bus_recovery_info[bus], info, sizeof(bus_recovery_info_T));
	return 0;   
}


int vsc_get_recovery_info(int bus, bus_recovery_info_T* info)
{
	if (bus >= BUS_COUNT) 
		return -1;

	memcpy(info, &m_bus_recovery_info[bus], sizeof(bus_recovery_info_T));
	return 0;   
}

/* Bus is considered hung if either clock or data are low for 150ms+ */
#define NUM_LINE_HUNG_CHECKS 75000
#define DELAY_BETWEEN_LINE_HUNG_CHECKS 2

int is_i2c_bus_hung(int bus)
{
	int NumTimesToCheck;

#ifdef SKIP_RECOVERY_OF_DEVICES_RECOVERED_FROM_GPIO
	if (SKIP_RECOVERY == skip_recovery(bus)) {
		return 0;
	}
#endif

#ifdef DEBUG
	printk(KERN_ERR "I2C%d: Checking for bus hang...", bus);
#endif
	/* If either clock or data is low */
	if ((LOW == get_sda_status(bus)) || (LOW == get_clk_status(bus))) {

		/* Check if SDA or CLK low for a specific period of time */
		NumTimesToCheck = NUM_LINE_HUNG_CHECKS;
		while (NumTimesToCheck) {
			touch_softlockup_watchdog();
			
#ifdef ARBITRATION_SUPPORT
			if (bus == I2C_BUS_2) {  
				RESTART_SHORT_TIMER(RESTART_TIMER_VALUE);  /* Restart support fpga short timer */	
			}
#endif

			if ((HIGH == get_sda_status(bus)) && (HIGH == get_clk_status(bus)))
				break;
			else 
				NumTimesToCheck--;
			udelay(DELAY_BETWEEN_LINE_HUNG_CHECKS);
		}
    
		/* If SDA or CLK was low for too long, consider bus hung and return error */
		if (NumTimesToCheck == 0) {
#ifdef DEBUG
			printk(" clock or data line hung\n"); 
#endif
			return 1;
		}
	}

#ifdef DEBUG
	printk(" bus appears ok\n");
#endif

	return 0;        
}

void
reset_mux_via_gpio(int bus)
{
	Gpio_data_t	gpio_data;

	switch (bus) {
	
	    case 1 : /* Bus 1 is i2c MUX 5 */
            {
	        /* Set i2c mux reset line as output */
	        gpio_data.PinNum = GPIO_RST_I2C_MUX5;
	        gpio_data.data = 1;
	        vsc_setdir(g_gpio_dev, &gpio_data);

	        /* Reset i2c mux */
	        gpio_data.PinNum = GPIO_RST_I2C_MUX5;
	        gpio_data.data = 1;
	        vsc_writegpio(g_gpio_dev, &gpio_data);
	        udelay(1);
	        gpio_data.PinNum = GPIO_RST_I2C_MUX5;
	        gpio_data.data = 0;
	        vsc_writegpio(g_gpio_dev, &gpio_data);
	        udelay(1);
	        gpio_data.PinNum = GPIO_RST_I2C_MUX5;
	        gpio_data.data = 1;
	        vsc_writegpio(g_gpio_dev, &gpio_data);
	        udelay(5);

                break;
	    }
	
	    case 3 : /* Bus 3 is i2c MUX 3 */
            {
	        /* Set i2c mux reset line as output */
	        gpio_data.PinNum = GPIO_RST_I2C_MUX3;
	        gpio_data.data = 1;
	        vsc_setdir(g_gpio_dev, &gpio_data);

	        /* Reset i2c mux */
	        gpio_data.PinNum = GPIO_RST_I2C_MUX3;
	        gpio_data.data = 1;
	        vsc_writegpio(g_gpio_dev, &gpio_data);
	        udelay(1);
	        gpio_data.PinNum = GPIO_RST_I2C_MUX3;
	        gpio_data.data = 0;
	        vsc_writegpio(g_gpio_dev, &gpio_data);
	        udelay(1);
	        gpio_data.PinNum = GPIO_RST_I2C_MUX3;
	        gpio_data.data = 1;
	        vsc_writegpio(g_gpio_dev, &gpio_data);
	        udelay(5);

                break;
	    }
            default:
            {
                break;
            }
	}
	return;
}

